@@ -4,7 +4,7 @@
  * This file is part of Libgcrypt.
  *
  * Libgcrypt is free software; you can redistribute it and/or modify
- * it under the terms of the GNU Lesser general Public License as
+ * it under the terms of the GNU Lesser General Public License as
  * published by the Free Software Foundation; either version 2.1 of
  * the License, or (at your option) any later version.
  *
@@ -55,6 +55,14 @@
 #include "hash-common.h"
 
 
+/* Helper macro to force alignment to 64 bytes.  */
+#ifdef HAVE_GCC_ATTRIBUTE_ALIGNED
+# define ATTR_ALIGNED_64  __attribute__ ((aligned (64)))
+#else
+# define ATTR_ALIGNED_64
+#endif
+
+
 /* USE_ARM_NEON_ASM indicates whether to enable ARM NEON assembly code. */
 #undef USE_ARM_NEON_ASM
 #ifdef ENABLE_NEON_SUPPORT
@@ -72,6 +80,17 @@
 # define USE_ARM_ASM 1
 #endif
 
+/* USE_ARM64_SHA512 indicates whether to enable ARMv8 SHA512 extension assembly
+ * code. */
+#undef USE_ARM64_SHA512
+#ifdef ENABLE_ARM_CRYPTO_SUPPORT
+# if defined(__AARCH64EL__) \
+       && defined(HAVE_COMPATIBLE_GCC_AARCH64_PLATFORM_AS) \
+       && defined(HAVE_GCC_INLINE_ASM_AARCH64_SHA3_SHA512_SM3_SM4)
+#  define USE_ARM64_SHA512 1
+# endif
+#endif
+
 
 /* USE_SSSE3 indicates whether to compile with Intel SSSE3 code. */
 #undef USE_SSSE3
@@ -104,6 +123,16 @@
 #endif
 
 
+/* USE_AVX512 indicates whether to compile with Intel AVX512 code. */
+#undef USE_AVX512
+#if defined(__x86_64__) && defined(HAVE_GCC_INLINE_ASM_AVX512) && \
+    defined(HAVE_INTEL_SYNTAX_PLATFORM_AS) && \
+    (defined(HAVE_COMPATIBLE_GCC_AMD64_PLATFORM_AS) || \
+     defined(HAVE_COMPATIBLE_GCC_WIN64_PLATFORM_AS))
+# define USE_AVX512 1
+#endif
+
+
 /* USE_SSSE3_I386 indicates whether to compile with Intel SSSE3/i386 code. */
 #undef USE_SSSE3_I386
 #if defined(__i386__) && SIZEOF_UNSIGNED_LONG == 4 && __GNUC__ >= 4 && \
@@ -134,7 +163,7 @@
 
 typedef struct
 {
-  u64 h0, h1, h2, h3, h4, h5, h6, h7;
+  u64 h[8];
 } SHA512_STATE;
 
 typedef struct
@@ -142,13 +171,13 @@ typedef struct
   gcry_md_block_ctx_t bctx;
   SHA512_STATE state;
 #ifdef USE_S390X_CRYPTO
-  u64 final_len_msb, final_len_lsb; /* needs to be right after state.h7. */
+  u64 final_len_msb, final_len_lsb; /* needs to be right after state.h[7]. */
   int use_s390x_crypto;
 #endif
 } SHA512_CONTEXT;
 
 
-static const u64 k[] =
+static ATTR_ALIGNED_64 const u64 k[] =
   {
     U64_C(0x428a2f98d728ae22), U64_C(0x7137449123ef65cd),
     U64_C(0xb5c0fbcfec4d3b2f), U64_C(0xe9b5dba58189dbbc),
@@ -197,7 +226,8 @@ static const u64 k[] =
  * stack to store XMM6-XMM15 needed on Win64. */
 #undef ASM_FUNC_ABI
 #undef ASM_EXTRA_STACK
-#if defined(USE_SSSE3) || defined(USE_AVX) || defined(USE_AVX2)
+#if defined(USE_SSSE3) || defined(USE_AVX) || defined(USE_AVX2) \
+    || defined(USE_AVX512)
 # ifdef HAVE_COMPATIBLE_GCC_WIN64_PLATFORM_AS
 #  define ASM_FUNC_ABI __attribute__((sysv_abi))
 #  define ASM_EXTRA_STACK (10 * 16 + 4 * sizeof(void *))
@@ -208,6 +238,21 @@ static const u64 k[] =
 #endif
 
 
+#ifdef USE_ARM64_SHA512
+unsigned int _gcry_sha512_transform_armv8_ce (u64 state[8],
+                                              const unsigned char *data,
+                                              size_t num_blks,
+                                              const u64 k[]);
+
+static unsigned int
+do_sha512_transform_armv8_ce(void *ctx, const unsigned char *data,
+                             size_t nblks)
+{
+  SHA512_CONTEXT *hd = ctx;
+  return _gcry_sha512_transform_armv8_ce (hd->state.h, data, nblks, k);
+}
+#endif
+
 #ifdef USE_ARM_NEON_ASM
 unsigned int _gcry_sha512_transform_armv7_neon (SHA512_STATE *hd,
                                                 const unsigned char *data,
@@ -232,8 +277,10 @@ do_sha512_transform_amd64_ssse3(void *ctx, const unsigned char *data,
                                 size_t nblks)
 {
   SHA512_CONTEXT *hd = ctx;
-  return _gcry_sha512_transform_amd64_ssse3 (data, &hd->state, nblks)
-         + ASM_EXTRA_STACK;
+  unsigned int burn;
+  burn = _gcry_sha512_transform_amd64_ssse3 (data, &hd->state, nblks);
+  burn += burn > 0 ? ASM_EXTRA_STACK : 0;
+  return burn;
 }
 #endif
 
@@ -247,8 +294,10 @@ do_sha512_transform_amd64_avx(void *ctx, const unsigned char *data,
                               size_t nblks)
 {
   SHA512_CONTEXT *hd = ctx;
-  return _gcry_sha512_transform_amd64_avx (data, &hd->state, nblks)
-         + ASM_EXTRA_STACK;
+  unsigned int burn;
+  burn = _gcry_sha512_transform_amd64_avx (data, &hd->state, nblks);
+  burn += burn > 0 ? ASM_EXTRA_STACK : 0;
+  return burn;
 }
 #endif
 
@@ -262,8 +311,27 @@ do_sha512_transform_amd64_avx2(void *ctx, const unsigned char *data,
                                size_t nblks)
 {
   SHA512_CONTEXT *hd = ctx;
-  return _gcry_sha512_transform_amd64_avx2 (data, &hd->state, nblks)
-         + ASM_EXTRA_STACK;
+  unsigned int burn;
+  burn = _gcry_sha512_transform_amd64_avx2 (data, &hd->state, nblks);
+  burn += burn > 0 ? ASM_EXTRA_STACK : 0;
+  return burn;
+}
+#endif
+
+#ifdef USE_AVX512
+unsigned int _gcry_sha512_transform_amd64_avx512(const void *input_data,
+						 void *state,
+						 size_t num_blks) ASM_FUNC_ABI;
+
+static unsigned int
+do_sha512_transform_amd64_avx512(void *ctx, const unsigned char *data,
+                                 size_t nblks)
+{
+  SHA512_CONTEXT *hd = ctx;
+  unsigned int burn;
+  burn = _gcry_sha512_transform_amd64_avx512 (data, &hd->state, nblks);
+  burn += burn > 0 ? ASM_EXTRA_STACK : 0;
+  return burn;
 }
 #endif
 
@@ -277,7 +345,7 @@ do_sha512_transform_i386_ssse3(void *ctx, const unsigned char *data,
 			       size_t nblks)
 {
   SHA512_CONTEXT *hd = ctx;
-  return _gcry_sha512_transform_i386_ssse3 (&hd->state.h0, data, nblks);
+  return _gcry_sha512_transform_i386_ssse3 (hd->state.h, data, nblks);
 }
 #endif
 
@@ -312,14 +380,14 @@ static unsigned int
 do_sha512_transform_ppc8(void *ctx, const unsigned char *data, size_t nblks)
 {
   SHA512_CONTEXT *hd = ctx;
-  return _gcry_sha512_transform_ppc8 (&hd->state.h0, data, nblks);
+  return _gcry_sha512_transform_ppc8 (hd->state.h, data, nblks);
 }
 
 static unsigned int
 do_sha512_transform_ppc9(void *ctx, const unsigned char *data, size_t nblks)
 {
   SHA512_CONTEXT *hd = ctx;
-  return _gcry_sha512_transform_ppc9 (&hd->state.h0, data, nblks);
+  return _gcry_sha512_transform_ppc9 (hd->state.h, data, nblks);
 }
 #endif
 
@@ -332,7 +400,7 @@ do_sha512_transform_s390x (void *ctx, const unsigned char *data, size_t nblks)
 {
   SHA512_CONTEXT *hd = ctx;
 
-  kimd_execute (KMID_FUNCTION_SHA512, &hd->state.h0, data, nblks * 128);
+  kimd_execute (KMID_FUNCTION_SHA512, hd->state.h, data, nblks * 128);
   return 0;
 }
 
@@ -343,18 +411,18 @@ do_sha512_final_s390x (void *ctx, const unsigned char *data, size_t datalen,
   SHA512_CONTEXT *hd = ctx;
 
   /* Make sure that 'final_len' is positioned at correct offset relative
-   * to 'state.h0'. This is because we are passing 'state.h0' pointer as start of
-   * parameter block to 'klmd' instruction. */
+   * to 'state.h[0]'. This is because we are passing 'state.h[0]' pointer as
+   * start of parameter block to 'klmd' instruction. */
 
   gcry_assert (offsetof (SHA512_CONTEXT, final_len_msb)
-	       - offsetof (SHA512_CONTEXT, state.h0) == 8 * sizeof(u64));
+	       - offsetof (SHA512_CONTEXT, state.h[0]) == 8 * sizeof(u64));
   gcry_assert (offsetof (SHA512_CONTEXT, final_len_lsb)
 	       - offsetof (SHA512_CONTEXT, final_len_msb) == 1 * sizeof(u64));
 
   hd->final_len_msb = len_msb;
   hd->final_len_lsb = len_lsb;
 
-  klmd_execute (KMID_FUNCTION_SHA512, &hd->state.h0, data, datalen);
+  klmd_execute (KMID_FUNCTION_SHA512, hd->state.h, data, datalen);
   return 0;
 }
 #endif
@@ -381,6 +449,10 @@ sha512_init_common (SHA512_CONTEXT *ctx, unsigned int flags)
   if ((features & HWF_ARM_NEON) != 0)
     ctx->bctx.bwrite = do_sha512_transform_armv7_neon;
 #endif
+#ifdef USE_ARM64_SHA512
+  if ((features & HWF_ARM_NEON) && (features & HWF_ARM_SHA512))
+    ctx->bctx.bwrite = do_sha512_transform_armv8_ce;
+#endif
 #ifdef USE_SSSE3
   if ((features & HWF_INTEL_SSSE3) != 0)
     ctx->bctx.bwrite = do_sha512_transform_amd64_ssse3;
@@ -393,6 +465,10 @@ sha512_init_common (SHA512_CONTEXT *ctx, unsigned int flags)
   if ((features & HWF_INTEL_AVX2) && (features & HWF_INTEL_BMI2))
     ctx->bctx.bwrite = do_sha512_transform_amd64_avx2;
 #endif
+#ifdef USE_AVX512
+  if ((features & HWF_INTEL_AVX512) && (features & HWF_INTEL_CPU))
+    ctx->bctx.bwrite = do_sha512_transform_amd64_avx512;
+#endif
 #ifdef USE_PPC_CRYPTO
   if ((features & HWF_PPC_VCRYPTO) != 0)
     ctx->bctx.bwrite = do_sha512_transform_ppc8;
@@ -425,14 +501,14 @@ sha512_init (void *context, unsigned int flags)
   SHA512_CONTEXT *ctx = context;
   SHA512_STATE *hd = &ctx->state;
 
-  hd->h0 = U64_C(0x6a09e667f3bcc908);
-  hd->h1 = U64_C(0xbb67ae8584caa73b);
-  hd->h2 = U64_C(0x3c6ef372fe94f82b);
-  hd->h3 = U64_C(0xa54ff53a5f1d36f1);
-  hd->h4 = U64_C(0x510e527fade682d1);
-  hd->h5 = U64_C(0x9b05688c2b3e6c1f);
-  hd->h6 = U64_C(0x1f83d9abfb41bd6b);
-  hd->h7 = U64_C(0x5be0cd19137e2179);
+  hd->h[0] = U64_C(0x6a09e667f3bcc908);
+  hd->h[1] = U64_C(0xbb67ae8584caa73b);
+  hd->h[2] = U64_C(0x3c6ef372fe94f82b);
+  hd->h[3] = U64_C(0xa54ff53a5f1d36f1);
+  hd->h[4] = U64_C(0x510e527fade682d1);
+  hd->h[5] = U64_C(0x9b05688c2b3e6c1f);
+  hd->h[6] = U64_C(0x1f83d9abfb41bd6b);
+  hd->h[7] = U64_C(0x5be0cd19137e2179);
 
   sha512_init_common (ctx, flags);
 }
@@ -443,14 +519,14 @@ sha384_init (void *context, unsigned int flags)
   SHA512_CONTEXT *ctx = context;
   SHA512_STATE *hd = &ctx->state;
 
-  hd->h0 = U64_C(0xcbbb9d5dc1059ed8);
-  hd->h1 = U64_C(0x629a292a367cd507);
-  hd->h2 = U64_C(0x9159015a3070dd17);
-  hd->h3 = U64_C(0x152fecd8f70e5939);
-  hd->h4 = U64_C(0x67332667ffc00b31);
-  hd->h5 = U64_C(0x8eb44a8768581511);
-  hd->h6 = U64_C(0xdb0c2e0d64f98fa7);
-  hd->h7 = U64_C(0x47b5481dbefa4fa4);
+  hd->h[0] = U64_C(0xcbbb9d5dc1059ed8);
+  hd->h[1] = U64_C(0x629a292a367cd507);
+  hd->h[2] = U64_C(0x9159015a3070dd17);
+  hd->h[3] = U64_C(0x152fecd8f70e5939);
+  hd->h[4] = U64_C(0x67332667ffc00b31);
+  hd->h[5] = U64_C(0x8eb44a8768581511);
+  hd->h[6] = U64_C(0xdb0c2e0d64f98fa7);
+  hd->h[7] = U64_C(0x47b5481dbefa4fa4);
 
   sha512_init_common (ctx, flags);
 }
@@ -462,14 +538,14 @@ sha512_256_init (void *context, unsigned int flags)
   SHA512_CONTEXT *ctx = context;
   SHA512_STATE *hd = &ctx->state;
 
-  hd->h0 = U64_C(0x22312194fc2bf72c);
-  hd->h1 = U64_C(0x9f555fa3c84c64c2);
-  hd->h2 = U64_C(0x2393b86b6f53b151);
-  hd->h3 = U64_C(0x963877195940eabd);
-  hd->h4 = U64_C(0x96283ee2a88effe3);
-  hd->h5 = U64_C(0xbe5e1e2553863992);
-  hd->h6 = U64_C(0x2b0199fc2c85b8aa);
-  hd->h7 = U64_C(0x0eb72ddc81c52ca2);
+  hd->h[0] = U64_C(0x22312194fc2bf72c);
+  hd->h[1] = U64_C(0x9f555fa3c84c64c2);
+  hd->h[2] = U64_C(0x2393b86b6f53b151);
+  hd->h[3] = U64_C(0x963877195940eabd);
+  hd->h[4] = U64_C(0x96283ee2a88effe3);
+  hd->h[5] = U64_C(0xbe5e1e2553863992);
+  hd->h[6] = U64_C(0x2b0199fc2c85b8aa);
+  hd->h[7] = U64_C(0x0eb72ddc81c52ca2);
 
   sha512_init_common (ctx, flags);
 }
@@ -481,14 +557,14 @@ sha512_224_init (void *context, unsigned int flags)
   SHA512_CONTEXT *ctx = context;
   SHA512_STATE *hd = &ctx->state;
 
-  hd->h0 = U64_C(0x8c3d37c819544da2);
-  hd->h1 = U64_C(0x73e1996689dcd4d6);
-  hd->h2 = U64_C(0x1dfab7ae32ff9c82);
-  hd->h3 = U64_C(0x679dd514582f9fcf);
-  hd->h4 = U64_C(0x0f6d2b697bd44da8);
-  hd->h5 = U64_C(0x77e36f7304c48942);
-  hd->h6 = U64_C(0x3f9d85a86a1d36c8);
-  hd->h7 = U64_C(0x1112e6ad91d692a1);
+  hd->h[0] = U64_C(0x8c3d37c819544da2);
+  hd->h[1] = U64_C(0x73e1996689dcd4d6);
+  hd->h[2] = U64_C(0x1dfab7ae32ff9c82);
+  hd->h[3] = U64_C(0x679dd514582f9fcf);
+  hd->h[4] = U64_C(0x0f6d2b697bd44da8);
+  hd->h[5] = U64_C(0x77e36f7304c48942);
+  hd->h[6] = U64_C(0x3f9d85a86a1d36c8);
+  hd->h[7] = U64_C(0x1112e6ad91d692a1);
 
   sha512_init_common (ctx, flags);
 }
@@ -543,14 +619,14 @@ do_transform_generic (void *context, const unsigned char *data, size_t nblks)
       int t;
 
       /* get values from the chaining vars */
-      a = hd->h0;
-      b = hd->h1;
-      c = hd->h2;
-      d = hd->h3;
-      e = hd->h4;
-      f = hd->h5;
-      g = hd->h6;
-      h = hd->h7;
+      a = hd->h[0];
+      b = hd->h[1];
+      c = hd->h[2];
+      d = hd->h[3];
+      e = hd->h[4];
+      f = hd->h[5];
+      g = hd->h[6];
+      h = hd->h[7];
 
       for ( t = 0; t < 16; t++ )
         w[t] = buf_get_be64(data + t * 8);
@@ -749,14 +825,14 @@ do_transform_generic (void *context, const unsigned char *data, size_t nblks)
         }
 
       /* Update chaining vars.  */
-      hd->h0 += a;
-      hd->h1 += b;
-      hd->h2 += c;
-      hd->h3 += d;
-      hd->h4 += e;
-      hd->h5 += f;
-      hd->h6 += g;
-      hd->h7 += h;
+      hd->h[0] += a;
+      hd->h[1] += b;
+      hd->h[2] += c;
+      hd->h[3] += d;
+      hd->h[4] += e;
+      hd->h[5] += f;
+      hd->h[6] += g;
+      hd->h[7] += h;
 
       data += 128;
     }
@@ -818,7 +894,6 @@ sha512_final (void *context)
 	  hd->bctx.buf[hd->bctx.count++] = 0x80;  /* pad */
 	  if (hd->bctx.count < 112)
 	    memset (&hd->bctx.buf[hd->bctx.count], 0, 112 - hd->bctx.count);
-	  hd->bctx.count = 112;
 	}
       else
 	{
@@ -837,7 +912,7 @@ sha512_final (void *context)
     }
 
   p = hd->bctx.buf;
-#define X(a) do { buf_put_be64(p, hd->state.h##a); p += 8; } while (0)
+#define X(a) do { buf_put_be64(p, hd->state.h[a]); p += 8; } while (0)
   X (0);
   X (1);
   X (2);
@@ -850,6 +925,8 @@ sha512_final (void *context)
   X (7);
 #undef X
 
+  hd->bctx.count = 0;
+
   _gcry_burn_stack (burn);
 }
 
@@ -861,25 +938,15 @@ sha512_read (void *context)
 }
 
 
-/* Shortcut functions which puts the hash value of the supplied buffer
+/* Shortcut functions which puts the hash value of the supplied buffer iov
  * into outbuf which must have a size of 64 bytes.  */
-void
-_gcry_sha512_hash_buffer (void *outbuf, const void *buffer, size_t length)
+static void
+_gcry_sha512_hash_buffers (void *outbuf, size_t nbytes,
+			   const gcry_buffer_t *iov, int iovcnt)
 {
   SHA512_CONTEXT hd;
 
-  sha512_init (&hd, 0);
-  _gcry_md_block_write (&hd, buffer, length);
-  sha512_final (&hd);
-  memcpy (outbuf, hd.bctx.buf, 64);
-}
-
-
-/* Variant of the above shortcut function using multiple buffers.  */
-void
-_gcry_sha512_hash_buffers (void *outbuf, const gcry_buffer_t *iov, int iovcnt)
-{
-  SHA512_CONTEXT hd;
+  (void)nbytes;
 
   sha512_init (&hd, 0);
   for (;iovcnt > 0; iov++, iovcnt--)
@@ -891,25 +958,15 @@ _gcry_sha512_hash_buffers (void *outbuf, const gcry_buffer_t *iov, int iovcnt)
 
 
 
-/* Shortcut functions which puts the hash value of the supplied buffer
+/* Shortcut functions which puts the hash value of the supplied buffer iov
  * into outbuf which must have a size of 48 bytes.  */
 static void
-_gcry_sha384_hash_buffer (void *outbuf, const void *buffer, size_t length)
+_gcry_sha384_hash_buffers (void *outbuf, size_t nbytes,
+			   const gcry_buffer_t *iov, int iovcnt)
 {
   SHA512_CONTEXT hd;
 
-  sha384_init (&hd, 0);
-  _gcry_md_block_write (&hd, buffer, length);
-  sha512_final (&hd);
-  memcpy (outbuf, hd.bctx.buf, 48);
-}
-
-
-/* Variant of the above shortcut function using multiple buffers.  */
-static void
-_gcry_sha384_hash_buffers (void *outbuf, const gcry_buffer_t *iov, int iovcnt)
-{
-  SHA512_CONTEXT hd;
+  (void)nbytes;
 
   sha384_init (&hd, 0);
   for (;iovcnt > 0; iov++, iovcnt--)
@@ -921,26 +978,15 @@ _gcry_sha384_hash_buffers (void *outbuf, const gcry_buffer_t *iov, int iovcnt)
 
 
 
-/* Shortcut functions which puts the hash value of the supplied buffer
+/* Shortcut functions which puts the hash value of the supplied buffer iov
  * into outbuf which must have a size of 32 bytes.  */
 static void
-_gcry_sha512_256_hash_buffer (void *outbuf, const void *buffer, size_t length)
+_gcry_sha512_256_hash_buffers (void *outbuf, size_t nbytes,
+			       const gcry_buffer_t *iov, int iovcnt)
 {
   SHA512_CONTEXT hd;
 
-  sha512_256_init (&hd, 0);
-  _gcry_md_block_write (&hd, buffer, length);
-  sha512_final (&hd);
-  memcpy (outbuf, hd.bctx.buf, 32);
-}
-
-
-/* Variant of the above shortcut function using multiple buffers.  */
-static void
-_gcry_sha512_256_hash_buffers (void *outbuf, const gcry_buffer_t *iov,
-			       int iovcnt)
-{
-  SHA512_CONTEXT hd;
+  (void)nbytes;
 
   sha512_256_init (&hd, 0);
   for (;iovcnt > 0; iov++, iovcnt--)
@@ -952,26 +998,15 @@ _gcry_sha512_256_hash_buffers (void *outbuf, const gcry_buffer_t *iov,
 
 
 
-/* Shortcut functions which puts the hash value of the supplied buffer
+/* Shortcut functions which puts the hash value of the supplied buffer iov
  * into outbuf which must have a size of 28 bytes.  */
 static void
-_gcry_sha512_224_hash_buffer (void *outbuf, const void *buffer, size_t length)
+_gcry_sha512_224_hash_buffers (void *outbuf, size_t nbytes,
+			       const gcry_buffer_t *iov, int iovcnt)
 {
   SHA512_CONTEXT hd;
 
-  sha512_224_init (&hd, 0);
-  _gcry_md_block_write (&hd, buffer, length);
-  sha512_final (&hd);
-  memcpy (outbuf, hd.bctx.buf, 28);
-}
-
-
-/* Variant of the above shortcut function using multiple buffers.  */
-static void
-_gcry_sha512_224_hash_buffers (void *outbuf, const gcry_buffer_t *iov,
-			       int iovcnt)
-{
-  SHA512_CONTEXT hd;
+  (void)nbytes;
 
   sha512_224_init (&hd, 0);
   for (;iovcnt > 0; iov++, iovcnt--)
@@ -1219,41 +1254,43 @@ run_selftests (int algo, int extended, selftest_report_func_t report)
 
 
 
-static byte sha512_asn[] =	/* Object ID is 2.16.840.1.101.3.4.2.3 */
+static const byte sha512_asn[] =	/* Object ID is 2.16.840.1.101.3.4.2.3 */
   {
     0x30, 0x51, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86,
     0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x03, 0x05,
     0x00, 0x04, 0x40
   };
 
-static gcry_md_oid_spec_t oid_spec_sha512[] =
+static const gcry_md_oid_spec_t oid_spec_sha512[] =
   {
     { "2.16.840.1.101.3.4.2.3" },
 
     /* PKCS#1 sha512WithRSAEncryption */
     { "1.2.840.113549.1.1.13" },
+    /* ANSI X9.62  ecdsaWithSHA512 */
+    { "1.2.840.10045.4.3.4" },
 
     { NULL }
   };
 
-gcry_md_spec_t _gcry_digest_spec_sha512 =
+const gcry_md_spec_t _gcry_digest_spec_sha512 =
   {
     GCRY_MD_SHA512, {0, 1},
     "SHA512", sha512_asn, DIM (sha512_asn), oid_spec_sha512, 64,
     sha512_init, _gcry_md_block_write, sha512_final, sha512_read, NULL,
-    _gcry_sha512_hash_buffer, _gcry_sha512_hash_buffers,
+    _gcry_sha512_hash_buffers,
     sizeof (SHA512_CONTEXT),
     run_selftests
   };
 
-static byte sha384_asn[] =	/* Object ID is 2.16.840.1.101.3.4.2.2 */
+static const byte sha384_asn[] =	/* Object ID is 2.16.840.1.101.3.4.2.2 */
   {
     0x30, 0x41, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86,
     0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x02, 0x05,
     0x00, 0x04, 0x30
   };
 
-static gcry_md_oid_spec_t oid_spec_sha384[] =
+static const gcry_md_oid_spec_t oid_spec_sha384[] =
   {
     { "2.16.840.1.101.3.4.2.2" },
 
@@ -1263,53 +1300,66 @@ static gcry_md_oid_spec_t oid_spec_sha384[] =
     /* SHA384WithECDSA: RFC 7427 (A.3.3.) */
     { "1.2.840.10045.4.3.3" },
 
+    /* ANSI X9.62  ecdsaWithSHA384 */
+    { "1.2.840.10045.4.3.3" },
+
     { NULL },
   };
 
-gcry_md_spec_t _gcry_digest_spec_sha384 =
+const gcry_md_spec_t _gcry_digest_spec_sha384 =
   {
     GCRY_MD_SHA384, {0, 1},
     "SHA384", sha384_asn, DIM (sha384_asn), oid_spec_sha384, 48,
     sha384_init, _gcry_md_block_write, sha512_final, sha512_read, NULL,
-    _gcry_sha384_hash_buffer, _gcry_sha384_hash_buffers,
+    _gcry_sha384_hash_buffers,
     sizeof (SHA512_CONTEXT),
     run_selftests
   };
 
-static byte sha512_256_asn[] = { 0x30 };
+static const byte sha512_256_asn[] =
+  {
+    0x30, 0x31, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86,
+    0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x06, 0x05,
+    0x00, 0x04, 0x20
+  };
 
-static gcry_md_oid_spec_t oid_spec_sha512_256[] =
+static const gcry_md_oid_spec_t oid_spec_sha512_256[] =
   {
     { "2.16.840.1.101.3.4.2.6" },
 
     { NULL },
   };
 
-gcry_md_spec_t _gcry_digest_spec_sha512_256 =
+const gcry_md_spec_t _gcry_digest_spec_sha512_256 =
   {
     GCRY_MD_SHA512_256, {0, 1},
     "SHA512_256", sha512_256_asn, DIM (sha512_256_asn), oid_spec_sha512_256, 32,
     sha512_256_init, _gcry_md_block_write, sha512_final, sha512_read, NULL,
-    _gcry_sha512_256_hash_buffer, _gcry_sha512_256_hash_buffers,
+    _gcry_sha512_256_hash_buffers,
     sizeof (SHA512_CONTEXT),
     run_selftests
   };
 
-static byte sha512_224_asn[] = { 0x30 };
+static const byte sha512_224_asn[] =
+  {
+    0x30, 0x2d, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86,
+    0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x05, 0x05,
+    0x00, 0x04, 0x1c
+  };
 
-static gcry_md_oid_spec_t oid_spec_sha512_224[] =
+static const gcry_md_oid_spec_t oid_spec_sha512_224[] =
   {
     { "2.16.840.1.101.3.4.2.5" },
 
     { NULL },
   };
 
-gcry_md_spec_t _gcry_digest_spec_sha512_224 =
+const gcry_md_spec_t _gcry_digest_spec_sha512_224 =
   {
     GCRY_MD_SHA512_224, {0, 1},
     "SHA512_224", sha512_224_asn, DIM (sha512_224_asn), oid_spec_sha512_224, 28,
     sha512_224_init, _gcry_md_block_write, sha512_final, sha512_read, NULL,
-    _gcry_sha512_224_hash_buffer, _gcry_sha512_224_hash_buffers,
+    _gcry_sha512_224_hash_buffers,
     sizeof (SHA512_CONTEXT),
     run_selftests
   };
