@@ -7,7 +7,7 @@
  * This file is part of Libgcrypt.
  *
  * Libgcrypt is free software; you can redistribute it and/or modify
- * it under the terms of the GNU Lesser general Public License as
+ * it under the terms of the GNU Lesser General Public License as
  * published by the Free Software Foundation; either version 2.1 of
  * the License, or (at your option) any later version.
  *
@@ -102,6 +102,9 @@ global_init (void)
   if (!pre_syscall_func)
     gpgrt_get_syscall_clamp (&pre_syscall_func, &post_syscall_func);
 
+  /* Add a handler to be called after log_fatal and log_debug.  */
+  _gcry_set_gpgrt_post_log_handler ();
+
   /* See whether the system is in FIPS mode.  This needs to come as
      early as possible but after ATH has been initialized.  */
   _gcry_initialize_fips_mode (force_fips_mode);
@@ -307,12 +310,7 @@ print_config (const char *what, gpgrt_stream_t fp)
   if (!what || !strcmp (what, "cc"))
     {
       gpgrt_fprintf (fp, "cc:%d:%s:\n",
-#if GPGRT_VERSION_NUMBER >= 0x011b00 /* 1.27 */
-                     GPGRT_GCC_VERSION
-#else
-                     _GPG_ERR_GCC_VERSION /* Due to a bug in gpg-error.h.  */
-#endif
-                     ,
+                     GPGRT_GCC_VERSION,
 #ifdef __clang__
                      "clang:" __VERSION__
 #elif __GNUC__
@@ -356,6 +354,11 @@ print_config (const char *what, gpgrt_stream_t fp)
       gpgrt_fprintf (fp, "cpu-arch:"
 #if defined(HAVE_CPU_ARCH_X86)
                      "x86"
+#              ifdef __x86_64__
+                     ":amd64"
+#              else
+                     ":i386"
+#              endif
 #elif defined(HAVE_CPU_ARCH_ALPHA)
                      "alpha"
 #elif defined(HAVE_CPU_ARCH_SPARC)
@@ -375,6 +378,9 @@ print_config (const char *what, gpgrt_stream_t fp)
   if (!what || !strcmp (what, "mpi-asm"))
     gpgrt_fprintf (fp, "mpi-asm:%s:\n", _gcry_mpi_get_hw_config ());
 
+  if (!what || !strcmp (what, "mpi-pow"))
+    gpgrt_fprintf (fp, "mpi-powm:%s\n", _gcry_mpi_get_powm_config ());
+
   if (!what || !strcmp (what, "hwflist"))
     {
       unsigned int hwfeatures, afeature;
@@ -523,7 +529,7 @@ _gcry_vcontrol (enum gcry_ctl_cmds cmd, va_list arg_ptr)
   switch (cmd)
     {
     case GCRYCTL_ENABLE_M_GUARD:
-      _gcry_private_enable_m_guard ();
+      rc = GPG_ERR_NOT_SUPPORTED;
       break;
 
     case GCRYCTL_ENABLE_QUICK_RANDOM:
@@ -784,6 +790,17 @@ _gcry_vcontrol (enum gcry_ctl_cmds cmd, va_list arg_ptr)
       rc = _gcry_fips_run_selftests (1);
       break;
 
+    case GCRYCTL_FIPS_SERVICE_INDICATOR:
+      rc = _gcry_fips_indicator ();
+      break;
+
+    case GCRYCTL_FIPS_REJECT_NON_FIPS:
+      {
+        unsigned int flags = va_arg (arg_ptr, unsigned int);
+        _gcry_thread_context_set_reject (flags);
+      }
+      break;
+
     case GCRYCTL_FIPS_SERVICE_INDICATOR_CIPHER:
       /* Get FIPS Service Indicator for a given symmetric algorithm and
        * optional mode. Returns GPG_ERR_NO_ERROR if algorithm is allowed or
@@ -1059,20 +1076,6 @@ _gcry_is_secure (const void *a)
   return _gcry_private_is_secure (a);
 }
 
-void
-_gcry_check_heap( const void *a )
-{
-  (void)a;
-
-    /* FIXME: implement this*/
-#if 0
-    if( some_handler )
-	some_handler(a)
-    else
-	_gcry_private_check_heap(a)
-#endif
-}
-
 static void *
 _gcry_realloc_core (void *a, size_t n, int xhint)
 {
