@@ -186,24 +186,11 @@ check_oaep (void)
           err = gcry_pk_encrypt (&ciph, plain, pub_key);
           if (err)
             {
-              if (in_fips_mode)
-                {
-                  gcry_sexp_release (plain);
-                  plain = NULL;
-                  continue;
-                }
               show_sexp ("plain:\n", ciph);
               fail ("gcry_pk_encrypt failed: %s\n", gpg_strerror (err));
             }
           else
             {
-              if (in_fips_mode)
-                {
-                  fail ("The OAEP encryption unexpectedly worked in FIPS mode\n");
-                  gcry_sexp_release (plain);
-                  plain = NULL;
-                  continue;
-                }
               if (extract_cmp_data (ciph, "a", tbl[tno].m[mno].encr,
                                     tbl[tno].m[mno].desc))
                 {
@@ -217,19 +204,15 @@ check_oaep (void)
           plain = NULL;
 
           /* Now test the decryption.  */
-          seed = data_from_hex (tbl[tno].m[mno].seed, &seed_len);
           encr = data_from_hex (tbl[tno].m[mno].encr, &encr_len);
 
           err = gcry_sexp_build (&ciph, NULL,
                                  "(enc-val (flags oaep)(hash-algo sha1)"
-                                 "(random-override %b)"
                                  "(rsa (a %b)))",
-                                 (int)seed_len, seed,
                                  (int)encr_len, encr);
           if (err)
             die ("constructing cipher data failed: %s\n", gpg_strerror (err));
           gcry_free (encr);
-          gcry_free (seed);
 
           err = gcry_pk_decrypt (&plain, ciph, sec_key);
           if (err)
@@ -355,7 +338,6 @@ check_pss (void)
           sigtmpl = NULL;
 
           /* Now test the verification.  */
-          salt = data_from_hex (tbl[tno].m[mno].salt, &salt_len);
           sign = data_from_hex (tbl[tno].m[mno].sign, &sign_len);
 
           err = gcry_sexp_build (&sig, NULL,
@@ -365,14 +347,11 @@ check_pss (void)
             die ("constructing verify data failed: %s\n", gpg_strerror (err));
           err = gcry_sexp_build (&sigtmpl, NULL,
                                  "(data (flags pss)"
-                                 "(hash sha1 %b)"
-                                 "(random-override %b))",
-                                 20, mhash,
-                                 (int)salt_len, salt);
+                                 "(hash sha1 %b))",
+                                 20, mhash);
           if (err)
             die ("constructing verify tmpl failed: %s\n", gpg_strerror (err));
           gcry_free (sign);
-          gcry_free (salt);
 
           err = gcry_pk_verify (sig, sigtmpl, pub_key);
           if (err)
@@ -457,7 +436,7 @@ check_v15crypt (void)
           seed = data_from_hex (tbl[tno].m[mno].seed, &seed_len);
 
           err = gcry_sexp_build (&plain, NULL,
-                                 "(data (flags pkcs1)(hash-algo sha1)"
+                                 "(data (flags pkcs1)"
                                  "(value %b)(random-override %b))",
                                  (int)mesg_len, mesg,
                                  (int)seed_len, seed);
@@ -467,19 +446,7 @@ check_v15crypt (void)
           gcry_free (seed);
 
           err = gcry_pk_encrypt (&ciph, plain, pub_key);
-          if (in_fips_mode)
-            {
-              if (!err)
-                {
-                  fail ("gcry_pk_encrypt should have failed in FIPS mode:\n");
-                }
-              gcry_sexp_release (plain);
-              plain = NULL;
-              gcry_sexp_release (ciph);
-              ciph = NULL;
-              continue;
-            }
-          else if (err)
+          if (err)
             {
               show_sexp ("plain:\n", ciph);
               fail ("gcry_pk_encrypt failed: %s\n", gpg_strerror (err));
@@ -499,19 +466,15 @@ check_v15crypt (void)
           plain = NULL;
 
           /* Now test the decryption.  */
-          seed = data_from_hex (tbl[tno].m[mno].seed, &seed_len);
           encr = data_from_hex (tbl[tno].m[mno].encr, &encr_len);
 
           err = gcry_sexp_build (&ciph, NULL,
-                                 "(enc-val (flags pkcs1)(hash-algo sha1)"
-                                 "(random-override %b)"
+                                 "(enc-val (flags pkcs1)"
                                  "(rsa (a %b)))",
-                                 (int)seed_len, seed,
                                  (int)encr_len, encr);
           if (err)
             die ("constructing cipher data failed: %s\n", gpg_strerror (err));
           gcry_free (encr);
-          gcry_free (seed);
 
           err = gcry_pk_decrypt (&plain, ciph, sec_key);
           if (err)
